/*
 *	Compile me with:
 *
 *		gcc -Wall -o synclient synclient.c -lXext -lX11
 */

/*
 * Copyright (c) 2007, 2008  Paul Vojta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 */

/*
 * To do:
 *	argument processing phase 2
 *	SetControls (here, synpt.h, and in the driver itself)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include <X11/Xlibint.h>
#include <X11/extensions/Xext.h>
#include <X11/extensions/extutil.h>
#include "synpt.h"

static	XExtensionInfo	_synpt_info_data;
static	XExtensionInfo	*synpt_info		= &_synpt_info_data;
static	char		*synpt_extension_name	= SynptExtensionName;

static	const char	*prog;		/* program name */

#define SynptCheckExtension(dpy, i, val) \
		XextCheckExtension(dpy, i, synpt_extension_name, val)

#define	Number(x)	(sizeof (x) / sizeof (*x))

/*
 *	Argument processing
 */

const char		*dispname	= NULL;		/* display name */
Bool			test		= False;	/* if test mode */

struct option {
	const char	*longname;
	char		shortname;
	const char	**addr;
};

static	const struct option	options[]	= {
		{"display",	'd',	&dispname},
		{"test",	't',	NULL},
		{"version",	'v',	NULL},
		{"help",	'h',	NULL}};


/*
 *	Variable names and types
 */

struct var {
	const char	*name;
	int		vtype;
	unsigned int	value;
};

#define	OPT_FLAG	0
#define	OPT_BYTE	1
#define	OPT_SHORT	2
#define	OPT_GEOM	3

#ifdef offsetof
# define o(field)	offsetof(SynptControlRec, field)
#else
# define o(field)	((size_t) &((SynptControlRec *)0)->field)
#endif

static	const struct var	vars[]		= {
		{"Enabled",		OPT_FLAG,	SYN_ENABLED},
		{"HighRate",		OPT_FLAG,	SYN_FAST},
		{"Emulate3Buttons",	OPT_FLAG,	SYN_EMU3},
		{"Emulate3Timeout",	OPT_SHORT,	o(time_emu3)},
		{"TapToClick",		OPT_FLAG,	SYN_TAPCLICK},
		{"TapAndDrag",		OPT_FLAG,	SYN_TAPDRAG},
		{"EdgeMotionUseAlways",	OPT_FLAG,	SYN_EDGEMOALWYS},
		{"LeftEdgeScroll",	OPT_FLAG,	SYN_L_EDGE_SCRL},
		{"RightEdgeScroll",	OPT_FLAG,	SYN_R_EDGE_SCRL},
		{"TopEdgeScroll",	OPT_FLAG,	SYN_T_EDGE_SCRL},
		{"BottomEdgeScroll",	OPT_FLAG,	SYN_B_EDGE_SCRL},
		{"HorizTwoFingerScroll",OPT_FLAG,	SYN_H_2FNG_SCRL},
		{"VertTwoFingerScroll",	OPT_FLAG,	SYN_V_2FNG_SCRL},
		{"MaxTapTime",		OPT_SHORT,	o(time_maxtap)},
		{"TapDuration",		OPT_SHORT,	o(time_tapdur)},
		{"HorizScrollDelta",	OPT_SHORT,	o(h_scroll_delta)},
		{"VertScrollDelta",	OPT_SHORT,	o(v_scroll_delta)},
		{"FingerHigh",		OPT_BYTE,	o(finger_high)},
		{"FingerLow",		OPT_BYTE,	o(finger_low)},
		{"EdgeMotionMinZ",	OPT_BYTE,	o(edge_min_z)},
		{"EdgeMotionMaxZ",	OPT_BYTE,	o(edge_max_z)},
		{"EdgeMotionMinSpeed",	OPT_BYTE,	o(edge_min_speed)},
		{"EdgeMotionMaxSpeed",	OPT_BYTE,	o(edge_max_speed)},
		{"TapButton1",		OPT_BYTE,	o(tap_button[0])},
		{"TapButton2",		OPT_BYTE,	o(tap_button[1])},
		{"TapButton3",		OPT_BYTE,	o(tap_button[2])},
		{"ULCornerButton",	OPT_BYTE,	o(ul_button)},
		{"URCornerButton",	OPT_BYTE,	o(ur_button)},
		{"LLCornerButton",	OPT_BYTE,	o(ll_button)},
		{"LRCornerButton",	OPT_BYTE,	o(lr_button)},
		{"MotionPosLeft",	OPT_GEOM,	o(mo_g.l)},
		{"MotionPosRight",	OPT_GEOM,	o(mo_g.r)},
		{"MotionPosTop",	OPT_GEOM,	o(mo_g.t)},
		{"MotionPosBottom",	OPT_GEOM,	o(mo_g.b)},
		{"ScrollPosLeft",	OPT_GEOM,	o(scrl_g.l)},
		{"ScrollPosRight",	OPT_GEOM,	o(scrl_g.r)},
		{"ScrollPosTop",	OPT_GEOM,	o(scrl_g.t)},
		{"ScrollPosBottom",	OPT_GEOM,	o(scrl_g.b)},
		{"CornerPosLeft",	OPT_GEOM,	o(cor_g.l)},
		{"CornerPosRight",	OPT_GEOM,	o(cor_g.r)},
		{"CornerPosTop",	OPT_GEOM,	o(cor_g.t)},
		{"CornerPosBottom",	OPT_GEOM,	o(cor_g.b)},
};


/*
 *	Private utility routines
 */

static int close_display(Display *dpy, XExtCodes *codes);

static XExtensionHooks synpt_extension_hooks = {
	/* create_gc */		NULL,
	/* copy_gc */		NULL,
	/* flush_gc */		NULL,
	/* free_gc */		NULL,
	/* create_font */	NULL,
	/* free_font */		NULL,
	/* close_display */	close_display,
	/* wire_to_event */	NULL,
	/* event_to_wire */	NULL,
	/* error */		NULL,
	/* error_string */	NULL,
};

static XEXT_GENERATE_FIND_DISPLAY(find_display, synpt_info,
		synpt_extension_name, &synpt_extension_hooks,
		SynptNumberEvents, NULL)

static XEXT_GENERATE_CLOSE_DISPLAY(close_display, synpt_info)


/*
 *	Public routines
 */

/* copy a bunch of things from DPMS.c [do I really need to do this?] */

int
SynptGetVersion(Display *dpy,
  unsigned char *protocol,
  unsigned char *hw_version,
  unsigned char *hw_subvers,
  unsigned short *cap_bits)
{
	XExtDisplayInfo			*info	= find_display(dpy);
	register xSynptGetVersionReq	*req;
	xSynptGetVersionReply		rep;

	SynptCheckExtension(dpy, info, -1);

	LockDisplay(dpy);
	GetReq(SynptGetVersion, req);
	req->reqType = info->codes->major_opcode;
	req->synptReqType = X_SynptGetVersion;
	if (!_XReply(dpy, (xReply *) &rep, 0, xTrue)) {
	    UnlockDisplay(dpy);
	    SyncHandle();
	    return -1;
	}
	UnlockDisplay(dpy);
	SyncHandle();
	*protocol = rep.protocol;
	*hw_version = rep.hw_version;
	*hw_subvers = rep.hw_subvers;
	*cap_bits = rep.cap_bits;
	return rep.version;
}

Bool
SynptGetOptions(Display *dpy, SynptControlPtr optp)
{
	XExtDisplayInfo		*info	= find_display(dpy);
	xSynptGetOptionsReq	*req;
	xSynptGetOptionsReply	rep;

	SynptCheckExtension(dpy, info, False);

	LockDisplay(dpy);
	GetReq(SynptGetOptions, req);
	req->reqType = info->codes->major_opcode;
	req->synptReqType = X_SynptGetOptions;
	if (!_XReply(dpy, (xReply *) &rep,
	  (sizeof (xSynptGetOptionsReply) - sizeof (xReply)) >> 2, xTrue)) {
	    UnlockDisplay(dpy);
	    SyncHandle();
	    return False;
	}
	UnlockDisplay(dpy);
	SyncHandle();
	*optp = rep.opt;
	return True;
}

Bool
SynptSetOptions(Display *dpy, SynptControlPtr optp)
{
	XExtDisplayInfo		*info	= find_display(dpy);
	xSynptSetOptionsReq	*req;

	SynptCheckExtension(dpy, info, False);

	LockDisplay(dpy);
	GetReq(SynptSetOptions, req);
	req->reqType = info->codes->major_opcode;
	req->synptReqType = X_SynptSetOptions;
	req->opt = *optp;

	UnlockDisplay(dpy);
	SyncHandle();

	return True;
}

void
show_option(const struct var *vp, SynptControlPtr optp)
{
	CARD8	c;
	CARD16	s;

	printf("%-20s", vp->name);
	switch (vp->vtype) {
	    case OPT_FLAG:
		puts(optp->flags & vp->value ? "on" : "off");
		break;
	    case OPT_BYTE:
		memcpy(&c, (char *) optp + vp->value, sizeof c);
		printf("%d\n", c);
		break;
	    case OPT_SHORT:
		memcpy(&s, (char *) optp + vp->value, sizeof s);
		printf("%d\n", s);
		break;
	    case OPT_GEOM:
		memcpy(&c, (char *) optp + vp->value, sizeof c);
		printf("%.03f\n", (double) c / 255);
	}
}

int
main(int argc, char **argv)
{
	Display		*dpy;
	XExtDisplayInfo	*info;
	SynptControlRec	opt;
	int		value;
	unsigned char	protocol;
	unsigned char	hw_version, hw_subvers;
	unsigned short	cap_bits;
	char		**arg, **argend;
	char		**narg = argv + 1;
	const struct var *vp;

	prog = strrchr(*argv, '/');
	if (prog != NULL) ++prog; else prog = *argv;

	/* Phase 1 of options processing.  Leave Parm=value arguments for
	 * later, since they might (in the distant future) depend on
	 * SynptGetVersion().  */

	argend = argv + argc;
	for (arg = argv + 1; arg < argend; ++arg) {
	    if (**arg != '-')
		*narg++ = *arg;
	    else {
		char			*p = *arg + 1;
		const struct option	*opt = options;
		const struct option	*opt0 = NULL;

		if (*p != '-' && p[1] == '\0') {	/* if 1-letter option */
		    for (opt = options; opt < options + Number(options); ++opt)
			if (*p == opt->shortname)
			    opt0 = opt;
		}
		else {
		    int len;

		    if (*p == '-') ++p;
		    len = strlen(p);
		    for (opt = options; opt < options + Number(options); ++opt)
			if (memcmp(p, opt->longname, len) == 0) {
			    /* Check for ambiguous option (can't happen yet) */
			    opt0 = opt;
			}
		}

		if (opt0 == NULL) {
		    fprintf(stderr, "%s: invalid option `%s'\n", prog, *arg);
		    return 1;
		}

		if (opt0->addr != NULL) {
		    if (++arg >= argend) {
			fprintf(stderr,
			  "%s: option `%s' requires an argument\n",
			  prog, arg[-1]);
			return 1;
		    }
		    *(opt0->addr) = *arg;
		}
		else
		    switch (opt0->shortname) {
			case 't':
			    test = True;
			    break;
			case 'v':
			    puts("Version 0.00.");
			    return 0;
			case 'h':
			    puts("Help text not yet written.");
			    return 0;
		    }
	    }
	}

	dpy = XOpenDisplay(dispname);
	if (dpy == NULL) {
	    fprintf(stderr, "%s: Cannot open display.\n", prog);
	    return 1;
	}

	info = find_display(dpy);
	if (!XextHasExtension(info)) {
	    fprintf(stderr, "%s: Extension \"%s\" missing on display \"%s\".\n",
	      prog, synpt_extension_name, dpy->display_name);
	    XCloseDisplay(dpy);
	    return 1;
	}

	if (test) {
	    XCloseDisplay(dpy);
	    return 0;
	}

	value = SynptGetVersion(dpy, &protocol, &hw_version, &hw_subvers,
	  &cap_bits);
	if (value != SynptVersion) {
	    fprintf(stderr,
	      "%s: Version mismatch: %d (server) vs. %d (client)\n", prog,
	      value, SynptVersion);
	    XCloseDisplay(dpy);
	    return 1;
	}

	if (argv + 1 >= narg) {
	    switch (protocol) {
		case SYN_PROTO_RAW:
		    printf("Synaptics touchpad, running under raw protocol\n"
		      "Version %d.%d, extended capability bits %04x\n\n",
		      hw_version, hw_subvers, cap_bits);
		    break;

		case SYN_PROTO_EVENT:
		    puts("Synaptics touchpad, running under event protocol\n");
		    break;

		case SYN_PROTO_PSM:
		    puts("Synaptics touchpad, running under PSM protocol\n");
		    break;

		default:
		    printf("Synaptics touchpad, unknown protocol %02x\n\n",
		      protocol);
	    }
	}

	if (!SynptGetOptions(dpy, &opt)) {
	    XCloseDisplay(dpy);
	    return 1;
	}

	if (argv + 1 < narg) {
	    const struct var **vlpp, **vlpp2;
	    Bool bad = False;
	    Bool changed = False;

	    vlpp = vlpp2 = malloc((narg - argv - 1) * sizeof *vlpp);
	    if (vlpp == NULL) {
		fprintf(stderr, "%s: malloc error\n", prog);
		XCloseDisplay(dpy);
		return 1;
	    }

	    for (arg = argv + 1; arg < narg; ++arg) {
		char *p;
		size_t len;
		const struct var *vp1;
		Bool bad1;

		p = strchr(*arg, '=');
		len = (p == NULL ? strlen(*arg) : p - *arg);
		vp1 = NULL;
		bad1 = False;
		for (vp = vars; vp < vars + Number(vars); ++vp) {
		    size_t len1 = strlen(vp->name);

		    if (len <= len1 && strncasecmp(*arg, vp->name, len) == 0) {
			if (len == len1) {
			    vp1 = vp;
			    bad1 = False;
			    break;
			}
			else {
			    if (vp1 != NULL) bad1 = True;
			    vp1 = vp;
			}
		    }
		}

		if (vp1 == NULL) {
		    fprintf(stderr, "%s: invalid variable %.*s\n", prog,
		      (int) len, *arg);
		    bad = True;
		}
		else if (bad1) {
		    fprintf(stderr, "%s: ambiguous variable %.*s\n", prog,
		      (int) len, *arg);
		    bad = True;
		}
		else {
		    if (p == NULL)
			*vlpp2++ = vp1;
		    else {
			++p;
			switch (vp1->vtype) {
			    case OPT_FLAG:
				if (strcmp(p, "0") == 0
				  || strcasecmp(p, "off") == 0
				  || strcasecmp(p, "f") == 0
				  || strcasecmp(p, "false") == 0
				  || strcasecmp(p, "n") == 0
				  || strcasecmp(p, "no") == 0)
				    opt.flags &= ~vp1->value;
				else if (strcmp(p, "1") == 0
				  || strcasecmp(p, "on") == 0
				  || strcasecmp(p, "t") == 0
				  || strcasecmp(p, "true") == 0
				  || strcasecmp(p, "y") == 0
				  || strcasecmp(p, "yes") == 0)
				    opt.flags |= vp1->value;
				else {
				    fprintf(stderr, "%s: invalid value `%s' "
				      "for boolean option `%s'\n", prog, p,
				      vp1->name);
				    bad = True;
				}
				break;

			    case OPT_BYTE:
			    case OPT_SHORT:
				if (!isdigit(*p)) {
				    fprintf(stderr, "%s: invalid value `%s' "
				      "for integer option `%s'\n", prog, p,
				      vp1->name);
				    bad = True;
				}
				else if (vp1->vtype == OPT_BYTE) {
				    CARD8 c = atoi(p);

				    memcpy((char *) &opt + vp1->value, &c,
				      sizeof c);
				}
				else {
				    CARD16 s = atoi(p);

				    memcpy((char *) &opt + vp1->value, &s,
				      sizeof s);
				}
				break;

			    case OPT_GEOM:
				{
				    double f;
				    CARD8 c;

				    f = atof(p);
				    if (f < 0 || f > 1) {
					fprintf(stderr, "%s: invalid value `%s'"
					  " for float option `%s'\n", prog, p,
					  vp1->name);
					bad = True;
					break;
				    }
				    c = f * 255 + 0.5;
				    memcpy((char *) &opt + vp1->value, &c,
				      sizeof c);
				}
			}

			changed = True;
		    }
		}
	    }
	    if (bad) {
		XCloseDisplay(dpy);
		return 1;
	    }

	    if (changed)
		if (!SynptSetOptions(dpy, &opt)) {
		    XCloseDisplay(dpy);
		    return 1;
		}

	    while (vlpp < vlpp2)
		show_option(*vlpp++, &opt);
	}
	else
	    for (vp = vars; vp < vars + Number(vars); ++vp)
		show_option(vp, &opt);

	XCloseDisplay(dpy);
	return 0;
}
